// Class: EnvPlotterGUI
//
// Author: Alyce Brady
//
// This class is based on the College Board's FishToolbar class,
// as allowed by the GNU General Public License.  FishToolbar is a
// black-box GUI class within the AP(r) CS Marine Biology Simulation
// case study (see www.collegeboard.com/ap/students/compsci).
//
// License Information:
//   This class is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation.
//
//   This class is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridLayout;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;

import java.lang.reflect.Method;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 *  Environment-Based Applications:<br>
 *
 *    The <code>EnvPlotterGUI</code> class provides the graphical user
 *    interface for a program that draws in an environment by placing
 *    color blocks in its cells.
 *
 *  @author Alyce Brady (based on code by Julie Zelenski)
 *  @version 1 January 2003
 **/
public class EnvPlotterGUI extends EnvFrameWithSlider
{
    // Specify the number of rows and columns in the environment.
    protected static final int ENV_ROWS = 5;
    protected static final int ENV_COLS = 5;

    // Specify the environment's background color and highlight color.
    protected static final Color BACKGROUND_COLOR = new Color(0, 155, 255);
    protected static final Color HIGHLIGHT_COLOR = Color.red;

    // Specify the minimum individual cell size.
    protected static final int MIN_CELL_SIZE = 10;

    // Instance Variables: Encapsulated data for EACH EnvPlotterGUI object
    protected ArrayList controlComponents = new ArrayList();
    protected CreateEnvDialog createEnvDialog = null;
    protected ColorChoiceDDMenu backgroundColorMenu, drawingColorMenu;
    protected EnvPlotter plotter;
    protected List drawingMethods;

  // constructor

    /** Constructs the graphical user interface for the Environment
     *  Plotter program.
     *    @param title frame title
     *    @param viewingWidth  the width of the viewing area
     *    @param viewingHeight the height of the viewing area
     *    @param maxDelayMsecs longest delay value for slider, in milliseconds
     *    @param minDelayMsecs shortest delay value for slider, in milliseconds
     **/
    public EnvPlotterGUI(String title, int viewingWidth, int viewingHeight,
                         int maxDelayMSecs, int minDelayMsecs)
    {
        // Build the list of drawing methods from the EnvPlotter class.
        getDrawingMethods();

        // Construct and initialize the GUI window.
        init(title, BACKGROUND_COLOR, viewingWidth, viewingHeight,
             MIN_CELL_SIZE, maxDelayMSecs, minDelayMsecs);

        // Construct a bounded environment in which to draw by coloring
        // cells.  Specify what object knows how to display color blocks,
        // then display the environment.
        newEnv();
        DisplayMap.associate("ColorBlock", new ColorBlockDisplay());
        showEnv();

        // Construct a plotter object that has methods for drawing
        // in an environment.
        plotter = new EnvPlotter((BoundedEnv)env, this);
        showEnv();
    }

    /** Builds a list of drawing methods that can be used by drawing
     *  buttons.
     **/
    protected void getDrawingMethods()
    {
        // Get all the methods from the EnvPlotter class, look through
        // them, and save all the drawing methods (void methods that
        // take a Color parameter) in a list.
        drawingMethods = new ArrayList();
        Method[] methods = EnvPlotter.class.getMethods();
        for ( int i = 0; i < methods.length; i++ )
        {
            Class retType = methods[i].getReturnType();
            Class[] paramTypes = methods[i].getParameterTypes();
            if ( retType.equals(void.class) &&
                 paramTypes.length == 1 && paramTypes[0].equals(Color.class) )
            {
                    drawingMethods.add(methods[i]);
            }
        }
    }

    /** Defines contents of main window panel.  Should be redefined in
     *  subclasses that require panels other than (or in addition to)
     *  a control panel and a display.
     *    @param backgroundColor color to paint background of environment
     *    @param viewingWidth  the width of the viewing area
     *    @param viewingHeight the height of the viewing area
     *    @param minCellSize   minimum environment cell side length
     **/
    protected void defineContent(Color bgColor,
                                int viewingWidth, int viewingHeight,
                                int minCellSize)
    {
        // Create a panel for the environment display and speed slider.
        super.defineContent(bgColor, viewingWidth, viewingHeight, minCellSize);

        // Create a control panel on the left.
        content.add(makeControlPanel(), BorderLayout.WEST);
    }


  // methods and nested classes for building the control panel

    /** Creates the control panel.
     **/
    protected JPanel makeControlPanel()
    {
        // Create a panel for the control buttons.
        JPanel p = new JPanel();
        p.setLayout(new BorderLayout());

        // Create a sub-panel for a button to create new environments
        // and color choosers for background and drawing color.
        JPanel subP = new JPanel();
        subP.setLayout(new BorderLayout());
        subP.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
        JButton newEnvButton = new JButton("Create New Environment");
        controlComponents.add(newEnvButton);
        newEnvButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
                { newEnv(); }});
        subP.add(newEnvButton, BorderLayout.NORTH);
        subP.add(makeColorChooserPanel(), BorderLayout.CENTER);
        p.add(subP, BorderLayout.NORTH);
        
        // Add drawing buttons.
        p.add(makeDrawingButtons(), BorderLayout.CENTER);
        
        // Add reset button.
        JButton resetButton = new JButton("Clear");
        controlComponents.add(resetButton);
        resetButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
                { resetEnv(); }});
        p.add(resetButton, BorderLayout.SOUTH);

        return p;
    }

    /** Creates a new environment.
     **/
    protected void newEnv()
    {
        // Ask user for environment dimensions, then construct an environment.
        if ( createEnvDialog == null )
            createEnvDialog = CreateEnvDialog.makeDimensionsDialog(this);
        Environment newEnv = createEnvDialog.showDialog();
        if ( env == null && newEnv == null )
            newEnv = new BoundedEnv(10, 10);
        if ( newEnv != null )
        {
            setEnv(newEnv);
            if ( plotter != null )
                plotter.configure((BoundedEnv)newEnv, this);
        }
    }

    /** Resets the environment.
     **/
    protected void resetEnv()
    {
        plotter.clearEnv();
    }

    /** Provides Color tools for choosing
     *  background color and drawing color.
     **/
    protected Component makeColorChooserPanel()
    {
        // Create a panel for the color choice menus.
        JPanel p = new JPanel();
        p.setLayout(new GridLayout(2, 2));
        p.setBorder(BorderFactory.createEmptyBorder(10, 4, 10, 4));

        p.add(new JLabel(" Background Color: "));
        backgroundColorMenu = new ColorChoiceDDMenu(ColorChoiceDDMenu.WHITE);
        controlComponents.add(backgroundColorMenu);
        backgroundColorMenu.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
                { setBackground(); }});
        p.add(backgroundColorMenu);
        
        p.add(new JLabel(" Drawing Color: "));
        drawingColorMenu = new ColorChoiceDDMenu(ColorChoiceDDMenu.RED);
        controlComponents.add(drawingColorMenu);
        p.add(drawingColorMenu);

        // Set the background color based on the selected choice.
        setBackground();

        return p;
    }

    /** Changes the background color.
     **/
    protected void setBackground()
    {
        display.setBackgroundColor(backgroundColorMenu.currentColor());
        showEnv();
    }

    /** Creates a panel containing the drawing buttons based
     *  on the methods provided in the EnvPlotter class.
     **/
    protected Component makeDrawingButtons()
    {
        // Create a panel for the drawing buttons and some space around them.
        JPanel enclosingPanel = new JPanel();
        enclosingPanel.setLayout(new BorderLayout());

        // Create a panel for the drawing buttons.
        JPanel p = new JPanel();
        p.setLayout(new GridLayout(0, 1));
        p.setBorder(BorderFactory.createCompoundBorder(
                BorderFactory.createTitledBorder("Drawing Buttons"), 
                BorderFactory.createEmptyBorder(4, 4, 4, 4)));

        Iterator iter = drawingMethods.iterator();
        while ( iter.hasNext() )
        {
            Method meth = (Method) iter.next();
            JButton aButton = new JButton(meth.getName());
            controlComponents.add(aButton);
            aButton.addActionListener(new AbstractAction(meth.getName()) {
                public void actionPerformed(ActionEvent e)
                    { draw((String) getValue(NAME)); }});
            p.add(aButton);
        }
        enclosingPanel.add(p, BorderLayout.NORTH);
        enclosingPanel.add(new JPanel(), BorderLayout.CENTER);

        return enclosingPanel;
    }

    /** Draws according to button pressed (called by action listener).
     *      @param buttonLabel  the label on the button pressed
     **/
    protected void draw(String buttonLabel)
    {
        Thread myThread = new Thread()
        {
            public void run ()
            {
                setControlComponentsEnabled(false);
                Method meth = getDrawingMethod(getName());
                Object[] methodArguments =
                    new Object[] { drawingColorMenu.currentColor() };
                try
                {   meth.invoke(plotter, methodArguments);  }
                catch (Exception e)
                {   throw new IllegalArgumentException("Can't invoke " +
                        meth.getName() + ": " + e);
                }
                showEnv();
                setControlComponentsEnabled(true);
            }
        };

        myThread.setName(buttonLabel);
        myThread.start();
    }

    /** Sets the enabled feature of the components in the control
     *  components list to <code>enabled</code>.
     *      @param enabled <code>true</code> if control components
     *                     should be enabled; <code>false</code> otherwise
     **/
    protected void setControlComponentsEnabled(boolean enabled)
    {
        Iterator iter = controlComponents.iterator();
        while ( iter.hasNext() )
        {
            JComponent component = (JComponent) iter.next();
            component.setEnabled(enabled);
        }
    }

    /** Gets the <code>EnvPlotter</code> drawing method associated
     *  with the specified button label.
     *      @param buttonLabel  the label on the button pressed
     **/
    protected Method getDrawingMethod(String buttonLabel)
    {
        Iterator iter = drawingMethods.iterator();
        while ( iter.hasNext() )
        {
            Method meth = (Method) iter.next();
            if ( meth.getName().equals(buttonLabel) )
                return meth;
        }
        throw new IllegalArgumentException("EnvPlotter has no " + buttonLabel +
                                           " method.");
    }

}
